/*
 * strdict.c
 *
 * string dictionary
 * hashed storage, with skiplist buckets
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sf_defs.h"
#include "xmalloc.h"
#include "error.h"
#include "strdict.h"
 
/* local prototypes */
static int rand_level(void); 
static int string_hash(string_dict d, char *str);
static int int_hash(int_dict d, int key);
static SkipNode *new_node(int level);
static SkipList *new_list(int (*compare)(), void (*assign)());
static void destroy_node(SkipNode *node);
static int intcmp(int *s1, int *s2);
static void assign_int(int **dest, int *src);

/***********************************************
 *
 * static int intcmp
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:09:22 1992
 *
 * DESCRIPTION:
 *  function for comparing two ints, given pointers
 *  to them (sort of like strcmp for ints)
 *
 * HISTORY:
 *  Tue Apr  7 21:09:22 1992 jz1j created
 *
 ***********************************************/
static int intcmp(int *s1, int *s2)
{
	int i1= (*s1);
	int i2= (*s2);

	return(i1-i2);
}

/***********************************************
 *
 * static int string_hash
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:09:46 1992
 *
 * DESCRIPTION:
 *  hashing function for strings, takes a string and
 *  returns a value such that 0 <= value < BUCKETS
 *
 * HISTORY:
 *  Tue Apr  7 21:09:46 1992 jz1j created
 *  Fri Jun 11 14:39:47 1993 jz1j variable # of buckets
 *
 ***********************************************/
static int string_hash(string_dict d, char *str)
{
#if 1
	int x;

	x = strlen(str)%d->buckets;
	return(x);
#else
	int x,r=0,l=strlen(str);

	if (d->buckets == 1)
		return(0);
	/* this should be replaced with something better */
    for(x=0;x<l;x++)
        r += str[x]^str[x+1];
    r = ((ABS(r)) % d->buckets);
    return(r);
#endif
}

/***********************************************
 *
 * static int int_hash
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:10:18 1992
 *
 * DESCRIPTION:
 *  hashing function for ints
 *  returns a value such that 0 <= value < BUCKETS
 *
 * HISTORY:
 *  Tue Apr  7 21:10:18 1992 jz1j created
 *  Fri Jun 11 14:43:33 1993 jz1j var # buckets
 *
 ***********************************************/
static int int_hash(int_dict d, int key)
{
	return(key%d->buckets);
}
 
/***********************************************
 *
 * static int rand_level
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:10:35 1992
 *
 * DESCRIPTION:
 *  returns a random level from 0-MAX_LEVEL
 *  higher levels are increasingly rare- this helps
 *  achieve ~(lg n) performance
 *
 * HISTORY:
 *  Tue Apr  7 21:10:35 1992 jz1j created
 *
 ***********************************************/
static int rand_level()
{
    int v;
    for(v=0;((random()&01) && v<MAX_LEVEL);v++);
    return(v);
}

/***********************************************
 *
 * static SkipNode *new_node
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:11:04 1992
 *
 * DESCRIPTION:
 *  creates a new SkipList node of the given level,
 *  allocates and initializes its next pointers
 *
 * HISTORY:
 *  Tue Apr  7 21:11:04 1992 jz1j created
 *
 ***********************************************/
static SkipNode *new_node(int level)
{
    SkipNode *new;
	int x;

	new = (SkipNode *)xmalloc(sizeof(SkipNode));
	new->forward = (SkipNode **)xmalloc((level+1)*sizeof(SkipNode *));
	new->key = NULL;
	for(x=0;x<=level;x++) {
		new->forward[x] = NULL;
	}

    return(new);
}

/***********************************************
 *
 * static SkipList *new_list
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:11:32 1992
 *
 * DESCRIPTION:
 *  creates a new SkipList that will use the given
 *  compare and assign functions
 *
 * HISTORY:
 *  Tue Apr  7 21:11:32 1992 jz1j created
 *
 ***********************************************/
static SkipList *new_list(int (*compare)(), void (*assign)())
{
    SkipList *new;
    SkipNode *empty;
    int x;
 
	new = (SkipList *)xmalloc(sizeof(SkipList));
    empty = new_node(0);
    empty->key = NULL;
    new->header = new_node(MAX_LEVEL);
    for(x=0;x<=MAX_LEVEL;x++)
        new->header->forward[x] = empty;
    new->level = 0;
	new->compare = compare;
	new->assign = assign;
    return(new);
}

/***********************************************
 *
 * caddr_t simple_dict_lookup
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:11:55 1992
 *
 * DESCRIPTION:
 *  looks a value (*key) up in a SkipList using
 *  its compare function, returns either the data
 *  associated with the key in the SkipList or NULL
 *  if there is none
 *
 * HISTORY:
 *  Tue Apr  7 21:11:55 1992 jz1j created
 *
 ***********************************************/
caddr_t simple_dict_lookup(simple_dict l, char *key)
{
    SkipNode *p;
    int x;

	if (!l) return(NULL);
	p = l->header;
	for(x=l->level;x>=0;x--)
		while(p->forward[x]->key && (l->compare(p->forward[x]->key,key) < 0))
			p = p->forward[x];
	p = p->forward[0];
	if (p->key && (!(l->compare(p->key,key))))
		return(p->data);
	return(NULL);
}

/***********************************************
 *
 * void simple_dict_set
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:13:14 1992
 *
 * DESCRIPTION:
 *  associate a value (data) with a key in the
 *  given SkipList
 *
 * HISTORY:
 *  Tue Apr  7 21:13:14 1992 jz1j created
 *
 ***********************************************/ 
void simple_dict_set(simple_dict list, char *key, caddr_t data)
{
	SkipNode *update[MAX_LEVEL+1];
	SkipNode *p = list->header;
	int x, newlevel;
	
    for(x=list->level;x>=0;x--) {
        while(p->forward[x]->key && (list->compare(p->forward[x]->key,key) < 0))
			p = p->forward[x];
        update[x] = p;
    }
    p = p->forward[0];
    if (p->key && (!(list->compare(p->key, key)))) {
		p->data = data;
	}
    else {
		if ((newlevel = rand_level()) > list->level) {
            for(x=list->level+1;x<=newlevel;x++)
				update[x] = list->header;
            list->level = newlevel;
        }
        p = new_node(newlevel);
        list->assign(&(p->key), key);
        p->data = data;
        for(x=0;x<=newlevel;x++) {
            p->forward[x] = update[x]->forward[x];
            update[x]->forward[x] = p;
        }
    }
}

/***********************************************
 *
 * void simple_dict_delete
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:13:35 1992
 *
 * DESCRIPTION:
 *  delete a value and its associated info
 *  from a skiplist
 *
 * HISTORY:
 *  Tue Apr  7 21:13:35 1992 jz1j created
 *
 ***********************************************/ 
void simple_dict_delete(simple_dict list, char *key)
{
    SkipNode *update[MAX_LEVEL+1];
    SkipNode *p = list->header;
    int x;
 
    for(x=list->level;x>=0;x--) {
        while(p->forward[x]->key && (list->compare(p->forward[x]->key,key) < 0))
            p = p->forward[x];
        update[x] = p;
    }    
    p = p->forward[0];
    if (list->compare(p->key,key)) return; /* already not in the list */
    for(x=list->level;x>=0;x--) {
        if (update[x]->forward[x] == p)
            update[x]->forward[x] = p->forward[x];
    }
    destroy_node(p);
    p = list->header; /* now fix up list->level if it should decrease */
    for(x=list->level;x>=0;x++) {
        if (p->forward[x] == list->header) list->level--;
        else return;
    }
}

/***********************************************
 *
 * void simple_dict_destroy
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:13:59 1992
 *
 * DESCRIPTION:
 *  destroy the given SkipList
 *
 * HISTORY:
 *  Tue Apr  7 21:13:59 1992 jz1j created
 *
 ***********************************************/
void simple_dict_destroy(simple_dict list)
{
    SkipNode *p,*q;
 
    for(p=list->header;p->forward[0]->key;p=q) {
        q = p->forward[0];
        destroy_node(p);
    }
    xfree(p->forward[0]);
    destroy_node(p);
}

/***********************************************
 *
 * static void destroy_node
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:15:28 1992
 *
 * DESCRIPTION:
 *  destroy a node and its associated data
 *  (note that the key is also destroyed)
 *
 * HISTORY:
 *  Tue Apr  7 21:15:28 1992 jz1j created
 *
 ***********************************************/ 
static void destroy_node(SkipNode *node)
{
	xfree(node->key);
    xfree(node->forward);
    xfree(node);
}

/***********************************************
 *
 * string_dict new_dictionary_nbuckets
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Fri Jun 11 14:49:28 1993
 *
 * DESCRIPTION:
 *  create a string dictionary with the given
 *  number of buckets
 *
 * HISTORY:
 *  Fri Jun 11 14:49:28 1993 jz1j created
 *
 ***********************************************/
string_dict new_dictionary_nbuckets(int buckets)
{
	string_dict new;
	int x;

	new = (string_dict)xmalloc(sizeof(struct t_Dictionary));
	new->buckets = buckets;
	for(x=0;x<new->buckets;x++)
		new->bucket[x] = new_list(strcmp, assign_string);
	return(new);
}

/***********************************************
 *
 * simple_string_dict new_simple_dict
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Fri Jun 11 15:39:21 1993
 *
 * DESCRIPTION:
 *  create a simple string dictionary (just
 *  a skiplist with strdict parms)
 *
 * HISTORY:
 *  Fri Jun 11 15:39:21 1993 jz1j created
 *
 ***********************************************/
simple_dict new_simple_dict()
{
	return(new_list(strcmp, assign_string));
}

/***********************************************
 *
 * string_dict new_dictionary
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:15:45 1992
 *
 * DESCRIPTION:
 *  returns a new string dictionary
 *  (string keys)
 *
 * HISTORY:
 *  Tue Apr  7 21:15:45 1992 jz1j created
 *  Fri Jun 11 14:43:40 1993 jz1j var # buckets
 *  Fri Jun 11 14:49:07 1993 jz1j became frontend for new_dictionary_nbuckets
 *
 ***********************************************/
string_dict new_dictionary()
{
	return(new_dictionary_nbuckets(DEFAULT_BUCKETS));
}

/***********************************************
 *
 * int_dict new_int_dictionary
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:15:54 1992
 *
 * DESCRIPTION:
 *  returns a new integer dictionary
 *  (integer keys)
 *
 * HISTORY:
 *  Tue Apr  7 21:15:54 1992 jz1j created
 *  Fri Jun 11 14:43:45 1993 jz1j var # buckets
 *
 ***********************************************/
int_dict new_int_dictionary()
{
	string_dict new;
	int x;

	new = (string_dict)xmalloc(sizeof(struct t_Dictionary));
	new->buckets = DEFAULT_BUCKETS;
	for(x=0;x<new->buckets;x++)
		new->bucket[x] = new_list(intcmp, assign_int);
	return(new);
}

/***********************************************
 *
 * void destroy_dict
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:16:14 1992
 *
 * DESCRIPTION:
 *  destroy a string dictionary
 *
 * HISTORY:
 *  Tue Apr  7 21:16:14 1992 jz1j created
 *  Fri Jun 11 14:43:50 1993 jz1j var # buckets
 *
 ***********************************************/
void destroy_dict(string_dict d)
{
	int x;

	for(x=0;x<d->buckets;x++)
		simple_dict_destroy(d->bucket[x]);
	xfree(d);
}

/***********************************************
 *
 * void destroy_int_dict
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:16:22 1992
 *
 * DESCRIPTION:
 *  destroy an integer dictionary
 *
 * HISTORY:
 *  Tue Apr  7 21:16:22 1992 jz1j created
 *  Fri Jun 11 14:43:58 1993 jz1j var # buckets
 *
 ***********************************************/
void destroy_int_dict(int_dict d)
{
	int x;

	for(x=0;x<d->buckets;x++)
		simple_dict_destroy(d->bucket[x]);
	xfree(d);
}

/***********************************************
 *
 * void set_dict
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:16:30 1992
 *
 * DESCRIPTION:
 *  set a value in a string dictionary
 *
 * HISTORY:
 *  Tue Apr  7 21:16:30 1992 jz1j created
 *
 ***********************************************/ 
void set_dict(string_dict d, char *key, caddr_t data)
{
	SkipList *l = d->bucket[string_hash(d, key)];

	simple_dict_set(l, key, data);
}

/***********************************************
 *
 * void set_int_dict
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:18:58 1992
 *
 * DESCRIPTION:
 *  set a value in an integer dictionary
 *  aren't these comments boring?
 *
 * HISTORY:
 *  Tue Apr  7 21:18:58 1992 jz1j created
 *
 ***********************************************/ 
void set_int_dict(int_dict d, int key, caddr_t data)
{
	SkipList *l = d->bucket[int_hash(d, key)];

	simple_dict_set(l, (char *)(&key), data);
}

/***********************************************
 *
 * void delete_from_dict
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:20:56 1992
 *
 * DESCRIPTION:
 *  delete key and its associated data from a
 *  string dictionary
 *
 * HISTORY:
 *  Tue Apr  7 21:20:56 1992 jz1j created
 *
 ***********************************************/
void delete_from_dict(string_dict d, char *key)
{
	SkipList *l = d->bucket[string_hash(d, key)];

	simple_dict_delete(l, key);
}

/***********************************************
 *
 * void delete_from_int_dict
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:21:13 1992
 *
 * DESCRIPTION:
 *  delete key and its associated data from an
 *  integer dictionary
 *
 * HISTORY:
 *  Tue Apr  7 21:21:13 1992 jz1j created
 *
 ***********************************************/ 
void delete_from_int_dict(int_dict d, int key)
{
	SkipList *l = d->bucket[int_hash(d, key)];

	simple_dict_delete(l, (char *)(&key));
}

/***********************************************
 *
 * caddr_t lookup_dict
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:21:28 1992
 *
 * DESCRIPTION:
 *  lookup and return the value associated with
 *  a key in a string dictionary
 *
 * HISTORY:
 *  Tue Apr  7 21:21:28 1992 jz1j created
 *  Fri Jun 11 14:47:52 1993 jz1j shortcut for 1 bucket dicts
 *
 ***********************************************/
caddr_t lookup_dict(string_dict d, char *key)
{
	SkipList *l = ((d->buckets > 1) ? d->bucket[string_hash(d, key)] : 0);

	return(simple_dict_lookup(l, key));
}

/***********************************************
 *
 * caddr_t lookup_int_dict
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:21:43 1992
 *
 * DESCRIPTION:
 *  lookup and return the value associated with
 *  a key in an integer dictionary
 *  does anyone ever read these comments?
 *  send mail to me if you really love this function
 *  heck, send mail if you even read this comment
 *
 * HISTORY:
 *  Tue Apr  7 21:21:43 1992 jz1j created
 *
 ***********************************************/
caddr_t lookup_int_dict(int_dict d, int key)
{
	SkipList *l = d->bucket[int_hash(d, key)];

	return(simple_dict_lookup(l, (char *)(&key)));
}

/***********************************************
 *
 * void traverse_dict
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:23:51 1992
 *
 * DESCRIPTION:
 *  traverse a string dictionary, calling proc on
 *  every entry in it (see skip_traverse for more
 *  info)
 *
 * HISTORY:
 *  Tue Apr  7 21:23:51 1992 jz1j created
 *  Sat May  9 23:09:27 1992 jz1j inlined skip_traverse
 *  Fri Jun 11 14:44:05 1993 jz1j var # buckets
 *
 ***********************************************/
void traverse_dict(string_dict d, void (*proc)(), caddr_t rock)
{
	SkipNode *r;
	int x;

	if (!d) {
		warn("NULL dict to traverse_dict");
		return;
	}
	for(x=0;x<d->buckets;x++) {
		r = d->bucket[x]->header;
		while(r) {
			if (r->key) proc(r->key, r->data, rock);
			r = r->forward[0];
		}
	}
}

/***********************************************
 *
 * void simple_dict_traverse
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Fri Jun 25 16:16:01 1993
 *
 * DESCRIPTION:
 *  traverse a simple dict, call
 *  proc(key, data, rock) on all entries
 *
 * HISTORY:
 *  Fri Jun 25 16:16:01 1993 jz1j created
 *
 ***********************************************/
void simple_dict_traverse(simple_dict l, void (*proc)(), caddr_t rock)
{
	SkipNode *r;

	r = l->header;
	while (r) {
		if (r->key)
			proc(r->key, r->data, rock);
		r = r->forward[0];
	}
}

/***********************************************
 *
 * void assign_string
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:25:31 1992
 *
 * DESCRIPTION:
 *  allocates space onto char *dest (note, it takes
 *  char **dest so it can do this) and stores the
 *  contents of src there
 *
 * HISTORY:
 *  Tue Apr  7 21:25:31 1992 jz1j created
 *
 ***********************************************/
void assign_string(char **dest, char *src)
{
	(*dest) = (char *)xmalloc(strlen(src)+1);
    strcpy((*dest), src);
}

/***********************************************
 *
 * static void assign_int
 *
 * AUTHOR: jz1j (James D. Zelenka) 
 * CREATED: Tue Apr  7 21:27:05 1992
 *
 * DESCRIPTION:
 *  just like assign_string for ints
 *  (c'mon, I'm tired of commenting functions)
 *
 * HISTORY:
 *  Tue Apr  7 21:27:05 1992 jz1j created
 *
 ***********************************************/
static void assign_int(int **dest, int *src)
{
	(*dest) = (int *)xmalloc(sizeof(int));
	(**dest) = (*src);
}


